
from django.db.models import Model
from django.forms.models import model_to_dict


def safe_model_to_dict(model):
    if isinstance(model, Model):
        return model_to_dict(model)
    return model


class DictInterface(object):
    """Wrapper for adding dict interface to object.

    Useful for models validation with standart validators
    """

    def __init__(self, model):
        self.model = model

    def __getattr__(self, name):
        if name is not 'model':
            return getattr(self.model, name)

    def __setattr__(self, name, value):
        if name is not 'model':
            return setattr(self.model, name, value)

    def __getitem__(self, name):
        return getattr(self.model, name)

    def __setitem__(self, name, value):
        return setattr(self.model, name, value)


def safe_model_dict_interface(model):
    if isinstance(model, Model):
        return DictInterface(model)
    return model



# built-in
from functools import partial       # noQA

# django
from django import forms

# project
from . import controllers
from . import exceptions
from . import parsers
from . import renderers
from . import validators

from .views import rule, ViewBase    # noQA


# shortcuts
c = controllers
e = exceptions
f = forms
p = parsers
r = renderers
v = validators



# built-in
from functools import partial
# django
from django.http import HttpResponseRedirect, JsonResponse, HttpResponse
from django.shortcuts import render
# project
from .exceptions import ValidationError


# PyYAML
try:
    import yaml as _yaml
except:
    _yaml = None


# BSON
try:
    import bson as _bson
except:
    _bson = None


# tablib
try:
    from tablib import Dataset as _Tablib
except:
    _Tablib = None # noQA


__all__ = [
    'Base',
    'Template', 'HTTP',
    'JSON', 'YAML', 'Tablib',
    'Redirect', 'Exception',
    ]


class Base(object):
    """Wrapper for using any function as renderer.
    """

    def __init__(self, renderer, content_name, request_name=None, names=None,
                 content=None, flat=False, **kwargs):
        """
        Args:
            - renderer (callable): wrapped function for rendering
            - content_name (str): key of data argument for renderer
            - request_name (str): key of Request argument for renderer
            - names (dict or None): dict of names. Keys:
                * data (default: "data"): name for data into content
                * errors (default: "errors"): name for errors into content
                * validator (default: None): name for validator into content.
                    If not setted, validator will not be passed into content.
            - content (dict): default params for content.
            - flat (bool): if True content contains only data or errors.
            - \**kwargs: some kwargs which will be passed to renderer.
        """
        self.renderer = renderer
        self.content_name = content_name
        self.request_name = request_name
        self.names = names or {}
        self.content = content
        self.flat = flat
        self.kwargs = kwargs

    def __call__(self, request=None, data=None, validator=None, status_code=None):
        # content
        if self.flat:
            content = data or (validator and validator.errors)
        else:
            content = {}
            # init data for content if passed
            if self.content:
                content.update(self.content)
            # main data for content
            if data:
                content[self.names.get('data', 'data')] = data
            if validator and 'validator' in self.names:
                content[self.names['validator']] = validator
            if validator and validator.errors:
                content[self.names.get('errors', 'errors')] = validator.errors

        # kwargs for rendering function
        params = self.kwargs.copy()
        params[self.content_name] = content
        if self.request_name:
            params[self.request_name] = request

        # rendering
        response = self.renderer(**params)
        if status_code:
            response.status_code = status_code
        return response


class BaseWithHTTP(Base):
    """Base class wrapped by HttpResponse
    """

    def set_http_kwargs(self, **kwargs):
        """Set kwargs for HttpResponse
        """
        self.http_kwargs = kwargs
        return self

    def __call__(self, status_code=None, **kwargs):
        http_kwargs = self.http_kwargs.copy()
        if status_code:
            http_kwargs['status_code'] = status_code
        content = super(BaseWithHTTP, self).__call__(**kwargs)
        return HttpResponse(content, **http_kwargs)


Template = partial(
    Base,
    renderer=render,
    content_name='context',
    request_name='request',
)
"""Serializer based on Django Templates
"""


class JSON(Base):
    """Serialize data into JSON
    """

    def __init__(self, flat=True, safe=False, **kwargs):
        """
        * Get all args of SerializerFactory.
        * Get all args of JsonResponse.
        * `flat` is True by default.
        * `safe` is False by default
        """
        super(JSON, self).__init__(
            renderer=JsonResponse,
            content_name='data',
            flat=flat,
            safe=safe,
            **kwargs)


class HTTP(object):
    """Render data by HttpResponse.

    `data` can be only `str` or `bytes` type.
    """

    def __init__(self, status_code=200, **kwargs):
        self.status_code = status_code
        self.kwargs = kwargs

    def __call__(self, request=None, data=None, validator=None, status_code=None):
        if not isinstance(data, (str, bytes, int, float, list)):
            t = type(data)
            if hasattr(t, '__name__'):
                t = t.__name__
            raise ValueError("Bad data format: {}".format(t))
        response = HttpResponse(data, **self.kwargs)
        response.status_code = status_code or self.status_code
        return response


class Redirect(object):
    """Redirect to URL

    URL can be passed by initialization or as data (str).
    """

    def __init__(self, url=None):
        self.url = url

    def __call__(self, data=None, **kwargs):
        url = self.url or data
        return HttpResponseRedirect(redirect_to=url)


class Exception(object): # noQA
    """Raise Exception

    I'm recommend use this renderer as `postr`.
    Raised exception can be handled by decorators or loggers.
    """

    def __init__(self, exception=ValidationError):
        self.exception = exception

    def __call__(self, request=None, data=None, validator=None, status_code=None):
        if validator and validator.errors:
            raise self.exception(validator.errors)
        else:
            raise self.exception(data)


class RESTFramework(BaseWithHTTP):
    """Wrapper for renderers from Django REST Framework
    """

    def __init__(self, renderer, flat=True, **kwargs):
        # initialize renderer
        if type(renderer) is type:
            renderer = renderer()
        self.http_kwargs = {}
        super(RESTFramework, self).__init__(
            renderer=renderer.render,
            content_name='data',
            flat=flat,
            **kwargs
        )


class YAML(BaseWithHTTP):
    """Render into YAML format by PyYAML
    """

    def __init__(self, flat=True, **kwargs):
        if not _yaml:
            raise ImportError('PyYAML is not installed yet')
        self.http_kwargs = {}
        super(YAML, self).__init__(
            renderer=_yaml.dump,
            content_name='data',
            flat=flat,
            **kwargs
        )


class BSON(BaseWithHTTP):
    """Render into BSON format by BSON package
    """

    def __init__(self, flat=True, **kwargs):
        if not _yaml:
            raise ImportError('BSON is not installed yet')
        self.http_kwargs = {}
        super(BSON, self).__init__(
            renderer=_bson.dumps,
            content_name='obj',
            flat=flat,
            **kwargs
        )


class Tablib(BaseWithHTTP):
    """Render into multiple formats by tablib
    """

    def __init__(self, ext, headers=None, **kwargs):
        if not _Tablib:
            raise ImportError('Tablib is not installed yet')
        self.http_kwargs = {}
        self.ext = ext
        self.headers = headers
        super(Tablib, self).__init__(
            renderer=self.render,
            content_name='data',
            flat=True,
            **kwargs
        )

    def render(self, data):
        dataset = _Tablib(*data, headers=self.headers)
        return dataset.export(self.ext)



# -*- coding: utf-8 -*-

# built-in
from collections import namedtuple
from functools import update_wrapper
# django
from django.views.generic import View
# project
from .exceptions import StatusCodeError, SubValidationError
from .parsers import Default as _DefaultParser


__all__ = ['rule', 'ViewBase']


_fields = ['d', 'p', 'prev', 'c', 'postv', 'prer', 'postr', 'r']
_Rule = namedtuple('Rule', _fields)


def _get_value(v, kwargs):
    if type(v) is str:
        return _get_value(kwargs[v], kwargs)
    else:
        return v


def rule(**kwargs):
    """Factory for _Rule objects

    * Any kwarg can contain str which point where function can get value
        for kwarg.
    * Some kwargs can contain None.

    Example:
        rule(
            d=[login_required, csrf_exempt],    # decorators
            prev=SomeDjangoForm,    # pre-validator
            c=some_controller,      # controller
            postv=None,             # post-validator
            # ^ here post-validator is missed
            prer='postr',           # renderer for pre-validator errors
            # ^ here `prer` point to `postr`
            postr='r',              # renderer for post-validator errors
            r=djburger.r.JSON(),    # renderer
        )

    Kwargs:
        - d (list): list of decorators
        - p (callable): parser. Parse request body.
            `djburger.p.Default` by default.
        - prev: pre-validator. Validate and clean user params
        - c (callable): controller
        - postv: post-validator. Validate and clean response
        - prer (callable): renderer for pre-validator errors
        - postr (callable): renderer for post-validator errors
        - r (callable): renderer for successfull response
    """
    # check required kwargs
    if 'c' not in kwargs:
        TypeError('Controller is required')
    if 'r' not in kwargs:
        TypeError('Renderer is required')
    # set default parser
    if 'p' not in kwargs:
        kwargs['p'] = _DefaultParser()
    # set 'r' as default for error-renderers
    for f in ('prer', 'postr'):
        if f not in kwargs:
            kwargs[f] = 'r'
    # set None as default for others
    for f in ('d', 'prev', 'postv'):
        if f not in kwargs:
            kwargs[f] = None

    for k, v in kwargs.items():
        kwargs[k] = _get_value(v, kwargs)
    return _Rule(**kwargs)


class ViewBase(View):
    """Base views for DjBurger usage
    """
    rules = None
    rule = None
    default_rule = None

    def dispatch(self, request, **kwargs):
        """Entrypoint for view

        1. Select rule from rules.
        2. Decorate view
        3. Call `validate` method.

        Args:
            - request (Request): Request object.
            - \**kwargs: kwargs from urls.py.

        Returns:
            - HttpResponse: django http response
        """
        self.method = request.method.lower()
        if self.rules and self.method in self.rules:
            self.rule = self.rules[self.method]
        elif self.default_rule:
            self.rule = self.default_rule
        else:
            raise NotImplementedError('Please, set rule or rules attr')

        base = self.validate_request

        # decorators
        if self.rule.d:
            for decorator in self.rule.d:
                base = decorator(base)
            # take possible attributes set by decorators like csrf_exempt
            base = update_wrapper(base, self.validate, assigned=())

        return base(request, **kwargs)

    def get_data(self, request):
        return self.rule.p(request)

    # pre-validator
    def validate_request(self, request, **kwargs):
        """
        1. Call `request_valid` method if validation is successfull or missed.
        2. Call `request_invalid` method otherwise.

        Args:
            - request (Request): Request object
            - \**kwargs: kwargs from urls.py.

        Returns:
            - HttpResponse: django http response
        """
        # data
        data = self.get_data(request)

        # no validator
        if not self.rule.prev:
            return self.request_valid(data, **kwargs)

        # validate
        validator = self.rule.prev(**self.get_validator_kwargs(data))
        try:
            is_valid = validator.is_valid()
        except StatusCodeError as e:
            is_valid = False
            status_code = e.status_code
        else:
            status_code = None
        if is_valid:
            return self.request_valid(validator.cleaned_data, **kwargs)
        else:
            return self.request_invalid(validator, status_code=status_code)

    # pre-validation error renderer
    def request_invalid(self, validator, status_code):
        """Return result of prer (renderer for pre-validator errors)

        Args:
            - validator: validator object with `errors` attr.
            - status_cdoe: status code for HTTP-response

        Returns:
            - HttpResponse: django http response
        """
        return self.rule.prer(
            request=self.request,
            validator=validator,
            status_code=status_code,
        )

    # controller
    def request_valid(self, data, **kwargs):
        """Call controller.

        Get response from controller and return result of validate_response
        method.

        Args:
            - data: cleaned and validated data from user.
            - \**kwargs: kwargs from urls.py.

        Returns:
            - HttpResponse: django http response
        """
        # get response from controller
        try:
            response = self.rule.c(self.request, data, **kwargs)
        except SubValidationError as e:
            validator = e.args[0]
            return self.subvalidation_invalid(validator)
        return self.validate_response(response)

    # post-validator
    def validate_response(self, response):
        """Validate response by postv (post-validator)

        1. Return make_response method result if post-validator is missed.
        2. Validate data by post-validator otherwise and call...
            * response_valid if validation is passed
            * or response_invalid otherwise.

        Args:
            - response: data from controller

        Returns:
            - HttpResponse: django http response
        """
        # no post-validator
        if not self.rule.postv:
            return self.make_response(data=response)

        # post-validation
        params = self.get_validator_kwargs(response)
        validator = self.rule.postv(**params)
        try:
            is_valid = validator.is_valid()
        except StatusCodeError as e:
            is_valid = False
            status_code = e.status_code
        else:
            status_code = None
        if is_valid:
            return self.response_valid(validator)
        else:
            return self.response_invalid(validator, status_code=status_code)

    # renderer for errors in subcontroller's validator
    def subvalidation_invalid(self, validator, status_code=200):
        return self.response_invalid(validator, status_code)

    # post-validation error renderer
    def response_invalid(self, validator, status_code):
        """Return result of postr (renderer for post-validation errors).

        Args:
            - validator: post-validator object with `errors` attr.
            - status_cdoe: status code for HTTP-response

        Returns:
            - HttpResponse: django http response
        """
        return self.rule.postr(
            request=self.request,
            validator=validator,
            status_code=status_code,
        )

    # successfull response renderer
    def response_valid(self, validator):
        """Return result of make_response.

        This method calls only if postv is not None.

        Args:
            - validator: validator object with `cleaned_data` attr.

        Returns:
            - HttpResponse: django http response
        """
        return self.make_response(validator.cleaned_data)

    def make_response(self, data):
        """Make response by renderer

        Args:
            - data: validated and cleaned data from controller

        Returns:
            - HttpResponse: django http response
        """
        return self.rule.r(request=self.request, data=data)

    # send request and data into validator
    def get_validator_kwargs(self, data):
        """Get kwargs for validators

        Args:
            - data: source data from user.

        Returns:
            - dict: kwargs for (post)validator
        """
        try:
            kwargs = super(ViewBase, self).get_form_kwargs()
        except AttributeError:
            kwargs = {}
        kwargs['request'] = self.request
        kwargs['data'] = data
        return kwargs



# -*- coding: utf-8 -*-
# django
from django.shortcuts import get_object_or_404
from django.views.generic import ListView
# project
from .exceptions import SubValidationError


__all__ = [
    'List', 'Info', 'Add', 'Edit', 'Delete',
    'ViewAsController',
    'pre', 'post', 'subcontroller',
]


class _ModelControllerMixin(object):
    def __init__(self, queryset=None, model=None):
        if queryset:
            self.q = queryset
        elif model:
            self.q = model._default_manager.all()
        else:
            raise ValueError("Queryset or model required.")


class List(ListView):
    """Controller based on Django ListView

    1. Get list of objects
    2. Filter list by validated data from user
    3. Optional pagination
    """

    def __init__(self, only_data=True, **kwargs):
        """Initialize controller in rule.

        Args:
            only_data (bool): return only filtered queryset if True,
                all context data otherwise. Use `only_data=False` with
                TemplateSerializerFactory.
            **kwargs: all arguments of ListView.
        """
        self.only_data = only_data
        super(List, self).__init__(**kwargs)

    def __call__(self, request, data, **kwargs):
        self.data = data
        return self.get(self, request, **kwargs)

    def get_queryset(self):
        q = super(List, self).get_queryset()
        return q.filter(**self.data)

    def get_context_data(self, **kwargs):
        context = super(List, self).get_context_data(**kwargs)
        # return only filtered queryset
        if self.only_data:
            return context['object_list']
        return context

    def render_to_response(self, context, **response_kwargs):
        return context


class Info(_ModelControllerMixin):
    """Return one object from queryset

    1. Return object filtered by params from URL kwargs  (like `pk` or `slug`)
        if url-pattern have kwargs.
    2. Return object from validated data if data have key `object`
        or data have only one key.
    3. Raise exception otherwise.
    """

    def __call__(self, request, data, **kwargs):
        if kwargs:
            return get_object_or_404(self.q, **kwargs)
        elif 'object' in data:
            return data['object']
        elif len(data) == 1:
            return data.pop()
        else:
            raise ValueError("Can't select object!")


class Add(object):
    """Controller for adding object with validated data.
    """

    def __init__(self, model):
        """Initialize controller with passed model.
        """
        self.model = model

    def __call__(self, request, data, **kwargs):
        return self.model._default_manager.create(**data)


class Edit(_ModelControllerMixin):
    """Controller for editing objects.

    1. Get object of initialized model by URL's kwargs.
    2. Set params from validated data.
    3. Update tuple into database.
    """

    def __call__(self, request, data, **kwargs):
        obj = get_object_or_404(self.q, **kwargs)
        obj.__dict__.update(data)
        obj.save(force_update=True)
        return obj


class Delete(_ModelControllerMixin):
    """Controller for deleting objects.

    Delete object filtered by URL's kwargs.
    """

    def __call__(self, request, data, **kwargs):
        obj = get_object_or_404(self.q, **kwargs)
        result = obj.delete()
        # hook for old Django versions
        if result is None:
            return 1
        return result[0]


class ViewAsController(object):
    """Allow use any django view as controller.

    1. For CBV with render_to_response method return context.
    2. Return rendered response otherwise.
    """

    def __init__(self, view, method=None):
        self.view = self.patch_view(view)
        self.method = method

    @staticmethod
    def patch_view(view):
        """Patch view for getting context instead of rendered response.
        """
        if hasattr(view, 'render_to_response'):
            view.render_to_response = lambda context, **kwargs: context
        return view

    def __call__(self, request, data, **kwargs):
        # set data to request
        method = self.method or request.method
        if method.lower() == 'get':
            request.GET = data
        else:
            request.POST = data

        # get response
        response = self.view(request=request, **kwargs)
        if hasattr(response, 'context'):
            return response.context
        elif hasattr(response, 'content'):
            return response.content
        else:
            return response


class pre(object): # noQA
    """Decorator for input data validation before subcontroller calling
    """
    def __init__(self, validator, **kwargs):
        self.validator = validator
        self.validator_kwargs = kwargs

    def __call__(self, controller):
        self.controller = controller
        return self._wrapper

    def _wrapper(self, data, request=None, **kwargs):
        validator = self.validator(
            data=data,
            request=request,
            **self.validator_kwargs,
        )
        if not validator.is_valid():
            raise SubValidationError(validator)
        return self.controller(
            data=validator.cleaned_data,
            request=request,
            **kwargs,
        )


class post(pre): # noQA
    """Decorator for output data validation before subcontroller calling
    """
    def _wrapper(self, data, request=None, **kwargs):
        result = self.controller(
            data=data,
            request=request,
            **kwargs,
        )
        validator = self.validator(
            data=result,
            request=request,
            **self.validator_kwargs,
        )
        if not validator.is_valid():
            raise SubValidationError(validator)
        return validator.cleaned_data


def subcontroller(c, prev=None, postv=None):
    """Constructor for subcontrollers
    If any validation failed, immediately raise SubValidationError.

    Kwargs:
        - prev (callable): validator
        - c (callable): controller
        - postv (callable): post-validator
    """
    if prev:
        c = pre(prev)(c)
    if postv:
        c = post(postv)(c)
    return c




# built-in
from functools import partial
from json import loads as _json


# BSON
try:
    from bson import loads as _bson
except:
    _bson = None


__all__ = ['Default', 'Base', 'JSON', 'BSON']


class Default(object):
    """Parse standart GET/POST query to dict
    """
    def __init__(self, method=None):
        self.method = method

    @staticmethod
    def _to_dict(query_dict):
        data = {}
        for k, v in query_dict.lists():
            data[k] = v[0] if len(v) == 1 else v
        return data

    def __call__(self, request):
        method = self.method or request.method.lower()
        if method == 'get':
            return self._to_dict(request.GET)
        else:
            return self._to_dict(request.POST)


class Base(object):
    """Allow use any callable object as parser
    """
    def __init__(self, parser, encoding='utf-8', **kwargs):
        if not parser:
            raise ImportError('Selected parser is not installed yet')
        self.parser = parser
        self.encoding = encoding
        self.kwargs = kwargs

    def __call__(self, request):
        body = request.body
        body = body.decode(self.encoding) if self.encoding else body
        return self.parser(body, **self.kwargs)


JSON = partial(Base, parser=_json)
BSON = partial(Base, parser=_bson, encoding=None)




from django.core.exceptions import ValidationError


class StatusCodeError(ValidationError):
    def __init__(self, status_code, msg, **kwargs):
        self.status_code = status_code
        self.msg = msg
        super(StatusCodeError, self).__init__(msg, **kwargs)


class SubValidationError(ValidationError):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup


setup(
    name     = 'djburger',
    version  = '0.7.0',

    author       = 'orsinium',
    author_email = 'master_fess@mail.ru',

    description  = 'Framework for views in big projects on Django.',
    long_description = open('README.md').read(),
    keywords     = 'djburger framework django contracts pre post validation',

    packages = ['djburger'],
    requires = ['python (>= 3.4)'],

    url          = 'https://github.com/orsinium/djburger',
    download_url = 'https://github.com/orsinium/djburger/tarball/master',

    license      = 'GNU Lesser General Public License v3.0',
    classifiers  = [
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python',
    ],
)
